/*Copyright 2011, 2013 Alexandros Chortaras

 This file is part of Rapid.

 Rapid is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Rapid is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Rapid.  If not, see <http://www.gnu.org/licenses/>.*/

package edu.ntua.ivml.common.dl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.io.StringDocumentSource;
import org.semanticweb.owlapi.io.StringDocumentTarget;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLDataProperty;
import org.semanticweb.owlapi.model.OWLEntity;
import org.semanticweb.owlapi.model.OWLNamedObject;
import org.semanticweb.owlapi.model.OWLObjectProperty;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyFormat;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.semanticweb.owlapi.model.OWLOntologyStorageException;
import org.semanticweb.owlapi.model.PrefixManager;
import org.semanticweb.owlapi.model.UnknownOWLOntologyException;
import org.semanticweb.owlapi.util.BidirectionalShortFormProvider;
import org.semanticweb.owlapi.vocab.PrefixOWLOntologyFormat;

import edu.ntua.ivml.common.lp.Atom;
import edu.ntua.ivml.common.lp.Clause;
import edu.ntua.ivml.common.lp.Predicate;
import edu.ntua.ivml.common.lp.Variable;

public class LoadedOntology implements BidirectionalShortFormProvider {

	private OWLOntology ontology;
	private OWLOntologyManager manager;
	
	private PrefixOWLOntologyFormat prefixFormat;
	
	private Set<String> conceptNames;
	private Set<String> roleNames;

	private Map<String, OWLEntity> entityMap;
	
	private boolean shortt;
	
	private IRI iri;

	public static LoadedOntology createFromString(String string) throws Exception {
		return createFromString(string, true);
	}
	
	public static LoadedOntology createFromString(String string, boolean shortt) throws Exception {
	
		OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
		OWLOntology ontology = manager.loadOntologyFromOntologyDocument(new StringDocumentSource(string));
		
		return new LoadedOntology(manager, ontology, shortt);
	}

	public static LoadedOntology createFromPath(String path) throws Exception {
		return createFromPath(path, true);
	}

	public static LoadedOntology createFromPath(String path, boolean shortt) throws Exception {
		return createFromIRI(IRI.create(path), shortt);
	}
	
	public static LoadedOntology createFromIRI(IRI iri) throws Exception {
		return 	createFromIRI(iri, true);
	}
	
	public static LoadedOntology createFromIRI(IRI iri, boolean shortt) throws Exception {
		OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
		OWLOntology ontology = manager.loadOntologyFromOntologyDocument(iri);
		
		LoadedOntology lo = new LoadedOntology(manager, ontology, shortt);
		lo.iri = iri;
		
		return lo;
	}

	public LoadedOntology(OWLOntologyManager manager, OWLOntology ontology) throws Exception {
		this(manager, ontology, true);
	}
	
	public LoadedOntology(OWLOntologyManager manager, OWLOntology ontology, boolean shortt) throws Exception {
		this.manager = manager;
		this.ontology = ontology;
		
		this.shortt = shortt;
		
		processPrefixFormat();
	}
	
	public IRI getIRI() {
		return iri;
	}
	
	private void processPrefixFormat() {
		prefixFormat =	manager.getOntologyFormat(ontology).asPrefixOWLOntologyFormat();

		entityMap = new HashMap<String, OWLEntity>();
		
		conceptNames = new HashSet<String>();
		roleNames = new HashSet<String>();
		
		for (OWLClass cl : ontology.getClassesInSignature()) {
			String c = handleIRI(cl);
			conceptNames.add(c);
			entityMap.put(c, cl);
		}

		for (OWLObjectProperty cl : ontology.getObjectPropertiesInSignature()) {
			String c = handleIRI(cl);
			roleNames.add(c);
			entityMap.put(c, cl);
		}
		
		for (OWLDataProperty cl : ontology.getDataPropertiesInSignature()) {
			entityMap.put(handleIRI(cl), cl);
		}

		entityMap.put(handleIRI(manager.getOWLDataFactory().getOWLNothing()), manager.getOWLDataFactory().getOWLNothing());
		entityMap.put(handleIRI(manager.getOWLDataFactory().getOWLThing()), manager.getOWLDataFactory().getOWLThing());
		
	}

	public PrefixManager getPrefixFormat() {
		return prefixFormat;
	}
	
	public void setShortForm(boolean s) {
		this.shortt = s;
	}
	
	public String getShortForm(OWLEntity arg0) {
		return handleIRI(arg0);
	}

	public Set<OWLEntity> getEntities(String arg0) {
		Set<OWLEntity> res = new HashSet<OWLEntity>();
		OWLEntity e = entityMap.get(arg0);
		if (e != null) {
			res.add(e);
		}
		
		return res;
	}

	public OWLEntity getEntity(String arg0) {
		return entityMap.get(arg0);
	}

	public Set<String> getShortForms() {
		return entityMap.keySet();
	}
	
	public OWLOntology getOntology() {
		return ontology;
	}
	
	public String handleIRI(OWLNamedObject obj) {

		if (shortt && obj.getIRI().getFragment() != null) {
			return obj.getIRI().getFragment().toString();
		} else {
			return obj.toString();
		}
	}

	public OWLOntologyManager getOntologyManager() {
		return manager;
	}

	public String toString(OWLOntologyFormat format) throws UnknownOWLOntologyException, OWLOntologyStorageException {
		StringDocumentTarget string = new StringDocumentTarget();

		manager.saveOntology(ontology, format, string);
	
		return string.toString();
	}
	
	public String getShortStringRepresentation(OWLEntity entity) {
		return handleIRI(entity);
	}
	
	public String getPhysicalURIPathForOntology(OWLOntology ont) {
		return manager.getOntologyDocumentIRI(ont).toURI().getPath();
	}

	public void dispose() {
		// TODO Auto-generated method stub
	}

	public Clause expand(Clause c) {
		ArrayList<Atom> newBody = new ArrayList<Atom>();
		for (Atom a : c.getBody()) {
			Predicate p  = a.getPredicate();
			Predicate np = p; 
			
			OWLEntity ent = getEntity(p.getName());
			if (ent != null) {
				np = new Predicate(ent.toString(), p.getArity());
			}
			
			newBody.add(new Atom(np, a.getArguments()));
		}

		Atom head = c.getHead();
		
		Predicate p  = head.getPredicate();
		Predicate np = p; 
		
		OWLEntity ent = getEntity(p.getName());
		if (ent != null) {
			np = new Predicate(ent.getIRI().toString(), p.getArity());
		}
		
		return new Clause(new Atom(np, head.getArguments()), newBody);
	}
	
	public ArrayList<Clause> getAtomClauses() {
		ArrayList<Clause> res = new ArrayList<>();
		
		for (OWLClass c : getOntology().getClassesInSignature()) {
			if (!c.isOWLThing()) {
				res.add(new Clause(new Atom(new Predicate(Predicate.QUERY_PREDICATE_NAME, 1), new Variable("x")), new Atom(new Predicate(handleIRI(c), 1), new Variable("x"))));
			}
		}
		
		for (OWLObjectProperty c : getOntology().getObjectPropertiesInSignature()) {
			res.add(new Clause(new Atom(new Predicate(Predicate.QUERY_PREDICATE_NAME, 2), new Variable("x"), new Variable("y")), new Atom(new Predicate(handleIRI(c), 2), new Variable("x"), new Variable("y"))));
			res.add(new Clause(new Atom(new Predicate(Predicate.QUERY_PREDICATE_NAME, 1), new Variable("x")), new Atom(new Predicate(handleIRI(c), 2), new Variable("x"), new Variable("y"))));
			res.add(new Clause(new Atom(new Predicate(Predicate.QUERY_PREDICATE_NAME, 1), new Variable("x")), new Atom(new Predicate(handleIRI(c), 2), new Variable("y"), new Variable("x"))));
		}
		
		return res;
	}
}
