/*Copyright 2011, 2013 Alexandros Chortaras

 This file is part of Rapid.

 Rapid is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Rapid is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Rapid.  If not, see <http://www.gnu.org/licenses/>.*/

package edu.ntua.ivml.qa.algorithm.rapid;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import edu.ntua.ivml.common.lp.Atom;
import edu.ntua.ivml.common.lp.Clause;
import edu.ntua.ivml.common.lp.Constant;
import edu.ntua.ivml.common.lp.Function;
import edu.ntua.ivml.common.lp.Predicate;
import edu.ntua.ivml.common.lp.Substitution;
import edu.ntua.ivml.common.lp.Term;
import edu.ntua.ivml.common.lp.Tuple;
import edu.ntua.ivml.common.lp.Variable;
import edu.ntua.ivml.common.utils.Counter;
import edu.ntua.ivml.common.utils.SetUtils;
import edu.ntua.ivml.qa.algorithm.PartialEngine;
import edu.ntua.ivml.qa.algorithm.ResolveEngine;
import edu.ntua.ivml.qa.algorithm.rapid.dllite.DLAtomUnfolder;
import edu.ntua.ivml.qa.rewriting.QueryCollector;

public abstract class Rapid extends PartialEngine implements ResolveEngine {

	public static String DUMMY_VAR_PREFIX = "v";
	public static String JOIN_VAR_PREFIX = "r";
	
	public static Variable UNBOUND_VARIABLE = new Variable(Variable.UNBOUND_PREFIX + "1");

	public static Variable[] fixedUnboundVars; 
	public static Variable[] extraUnboundVars;

	public static String FIXED_UNBOUND_PREFIX = "_f";
	public static String EXTRA_UNBOUND_PREFIX = "_w";

	public static ArrayList<Variable> dummyVars = new ArrayList<>();

	public static Set<Term> dummyTerms;

	static {
		fixedUnboundVars = new Variable[2];
		extraUnboundVars = new Variable[2];
		                
		for (int j = 0; j < 2; j++) {
			dummyVars.add(new Variable(DUMMY_VAR_PREFIX + (j + 1)));
			fixedUnboundVars[j] = new Variable(FIXED_UNBOUND_PREFIX + (j + 1));
			extraUnboundVars[j] = new Variable(EXTRA_UNBOUND_PREFIX + (j + 1));
		}			
		
		dummyTerms = new HashSet<Term>(Rapid.dummyVars);
	}

	public static boolean isJoinVariable(Variable v) {
		return v.getName().startsWith(JOIN_VAR_PREFIX);
	}

	public static boolean isCoreVariable(Variable v) {
		return !(isJoinVariable(v) || v.hasUnboundPrefix());
	}

	public static Substitution replaceUnboundVariablesByUnderscoreSubstitution(Clause c) {
		Substitution subst = new Substitution();
	
		Set<Variable> uVars = c.getUnboundVariables(); 
		
		for (Atom atom : c.getBody()) {
			Term[] terms = atom.getArguments().getTerms();
			for (int i = 0; i < terms.length; i++) {
				if (uVars.contains(terms[i])) {
					subst.put((Variable)terms[i], Rapid.fixedUnboundVars[i]);
				}
			}
		}
		
		return subst;
	}
	
	protected ClauseUnfolder clauseUnfolder;
	public AtomUnfolder atomUnfolder;
	protected Flattener flattener;
	
	protected boolean unfold;
	
	private Set<Clause> shrinkedClauses;
	
	protected Rapid(String name, boolean unfold) {
		super(name);
		
		this.unfold = unfold;
	}
	
	public AtomUnfolder getAtomUnfolder() {
		return atomUnfolder;
	}
	
	protected boolean preinit = false;
	
	public void preinit(Collection<Clause> atomClauses) {
		preinit = true;
		initialize();
		
		getAtomUnfolder().precompute(atomClauses);
	}
	
	public void execute() {
		compute();
	}
		
	protected void compute() {
		Counter rmark = new Counter(0);
		
		FlattenerExtra extra = new FlattenerExtra();
		
		ArrayList<MarkedClause> queue = new ArrayList<>();
		queue.add(new MarkedClause(query, rmark.getValue()));
		
		shrinkedClauses = new HashSet<>();
		shrinkedClauses.add(query);
		
		int k = 0;
		while (k < queue.size()) {
			Clause currentQuery = queue.get(k++).getClause(); 
			
			for (Variable v : SetUtils.difference(currentQuery.getBoundVariables(), currentQuery.getHead().getVariables())) {
				rmark.increase();
				
				for (MarkedClause c : resolve(currentQuery, v, rmark, extra)) {
					if (shrinkedClauses.add(c.getClause())) {
						queue.add(c);
					}
				}
			}
		}
		
		if (unfold) {
			int id = unfold(clauseUnfolder, qc, queue, extra);
			
			for (MarkedClause c : flattener.getExtraQueries(atomUnfolder, clauseUnfolder, extra)) {
				qc.add(id + c.getMark(), c.getClause());
			}
			
		} else {
			expand(strip(queue), extra);
		}
	}
	
	public static int unfold(ClauseUnfolder au, QueryCollector qc, Clause query, FlattenerExtra extra) {
		ArrayList<MarkedClause> cList = new ArrayList<MarkedClause>();
		cList.add(new MarkedClause(query, 0));
		
		return unfold(au, qc, cList, extra);
	}
	
	protected static int unfold(ClauseUnfolder cu, QueryCollector qc, ArrayList<MarkedClause> queryList, FlattenerExtra extra) {
		queryList = Clause.getNonRedundantClauses(queryList);
		
		int id = 0;
		
		int k = 0;
		while (k < queryList.size()) {
			MarkedClause mc = queryList.get(k++);
			
			Set<Atom> appendix = new HashSet<Atom>();
			int eIndex = -1;

			if (cu.getAtomUnfolder() instanceof DLAtomUnfolder) {
				while (k < queryList.size() && queryList.get(k).getMark() == mc.getMark()) {
					Clause nc = queryList.get(k++).getClause();
					if (eIndex == -1) {
						eIndex = nc.bodySize() - 1;
					} else if (eIndex > -1 && nc.bodySize() - 1 != eIndex) {
						throw new RuntimeException("ERROR!!!");
					} 
					
					appendix.add(nc.getBodyAtomAt(eIndex));
				}
			}
			
			ArrayList<Clause> last = new ArrayList<Clause>();

			Set<Variable> topBoundVars = mc.getClause().getBoundVariables();

			for (UnfoldedClause cp : cu.unfold(mc.getClause(), eIndex, appendix, extra)) {
				Clause up = cp.getClause().condenseWithOriginalVariables();

				if (cp.isPure() && up.getBoundVariables().containsAll(topBoundVars)) {
					qc.add(id, up, cp.isTop());
				} else {
					last.add(up);
				}
			}
			
			id++;
			for (Clause extraClause : last) {
				qc.add(id++, extraClause);
			}
		}
		
		return id;
	}
	
	public Set<Clause> getShrinkedClauses() {
		return shrinkedClauses;
	}

	public boolean isCGLLR() {
		return false;
	}
	
	protected abstract ArrayList<Clause> strip(ArrayList<MarkedClause> queryList);
	
	protected void expand(ArrayList<Clause> queries, FlattenerExtra extra) {
		Set<DatalogAtom> lsAtom = new HashSet<>();
		
		for (int k = 0; k < queries.size(); k++) {
			Clause c = queries.get(k);
				
			Set<Term> boundVars = c.getBoundTerms();
				
			ArrayList<Atom> newBody = new ArrayList<Atom>();
			
			for (Atom atom : c.getBody()) {
				DatalogAtom res = normalizeAtom(atom, boundVars);
				
				newBody.add(res.substAtom);
				lsAtom.add(res);
			}

			qc.add(0, new Clause(query.getHead(), newBody));
		}

		for (DatalogAtom da : lsAtom) {

			Atom[] aunf = atomUnfolder.datalogComputeUnfolding(da.rightAtom, Rapid.dummyVars, null, extra);

			for (Atom body : aunf) {
				if (!body.equals(da.rightAtom)) {
					qc.add(1, new Clause(da.leftAtom, body));
				}
			}
		}
		
		for (MarkedClause mc : flattener.getExtraQueries(atomUnfolder, clauseUnfolder, extra)) {
			Clause c  = mc.getClause();
			
			Set<Term> bTerms = c.getBoundTerms();
			
			ArrayList<Atom> newBody = new ArrayList<Atom>();
				
			for (Atom atom : c.getBody()) {
				DatalogAtom res = normalizeAtom(atom, bTerms);
					
				newBody.add(res.substAtom);
				lsAtom.add(res);
			}

			qc.add(1, new Clause(c.getHead(), newBody));
		}

		for (DatalogAtom da : lsAtom) {
			Clause ec = da.toDatalogClause();
			if (ec != null) {
				qc.add(1, ec);
			}
		}
	}
	
	public ArrayList<Clause> shallowResolve(Clause currentQuery, Set<Variable> vars) {
		ArrayList<Clause> res = new ArrayList<Clause>();

		for (Variable v : vars) {
			res.addAll(shallowResolve(currentQuery, v));
		}
		
		return res;
	}
	
	public ArrayList<Clause> shallowResolve(Clause currentQuery, Variable v) {
		ArrayList<Clause> res = new ArrayList<Clause>();
		
		Set<Variable> headVars = currentQuery.getHead().getVariables();
		if (headVars.contains(v)) {
			return res;
		}
		
		Set<Term> boundTerms = currentQuery.getBoundTerms();

		FunctionMap functionSet = null;
		Set<Variable> otherVariables = new HashSet<Variable>();
		Set<Constant> otherConstants = new HashSet<Constant>();

		int[] pattern = currentQuery.termIndexPattern(v);

		for (int k = 0; k < pattern.length; k++) {
			if (pattern[k] == 0) {
				continue;
			} else if (pattern[k] == 3) {
				return res;
			}
			
			Atom atom = currentQuery.getBodyAtomAt(k);

			AtomUnfolding aur = atomUnfolder.getShallowUnfolding(atom, boundTerms).getUnfolding();

			FunctionMap newFunctions = null;

			if (atom.isConcept()) {
				newFunctions = aur.getFunctionMap(0);
			} else {
				Term otherTerm = null;
			
				if (pattern[k] == 1) {
					newFunctions = aur.getFunctionMap(0);
					if (boundTerms.contains(atom.getArgument(1))) {
						otherTerm = atom.getArgument(1);
					}
				} else if (pattern[k] == 2) {
					newFunctions = aur.getFunctionMap(1);
					if (boundTerms.contains(atom.getArgument(0))) {
						otherTerm = atom.getArgument(0);
					}
				}
				
				if (otherTerm != null) {
					if (otherTerm.isVariable()) {
						otherVariables.add((Variable)otherTerm);
					} else  if (otherTerm.isConstant()) {
						otherConstants.add((Constant)otherTerm);
					}
				}
			}

			if (otherConstants.size() > 1) {
				return res;
			}

			if (functionSet == null) {
				functionSet = newFunctions.copy();
			} else {
				functionSet.intersection(newFunctions);
			}
			
			if (functionSet.isEmpty()) {
				return res;
			}
		}
	
		if (functionSet == null) {
			return res;
		}
		
		Term substTerm = null;
		Substitution subst = new Substitution();
		
		Set<Variable> allVariables = currentQuery.getVariables();

		if (otherConstants.size() == 0) {
			if (otherVariables.size() > 0) {
				Iterator<Variable> iter = otherVariables.iterator();
				substTerm = iter.next();
				while (!headVars.contains(substTerm) && iter.hasNext()) {
					substTerm = iter.next();
				}
			} else {
				substTerm = Variable.getNewVariable(Variable.UNBOUND_PREFIX, allVariables);
				allVariables.add((Variable)substTerm);
			}
		} else {
			substTerm = otherConstants.iterator().next();
		}

		for (Variable vv : otherVariables) {
			if (!vv.equals(substTerm)) {
				subst.put(vv, substTerm);
			}
		}

		ArrayList<Atom> baseBody = new ArrayList<Atom>();
		for (int k = 0; k < currentQuery.bodySize(); k++) {
			if (pattern[k] == 0) {
				baseBody.add(currentQuery.getBodyAtomAt(k).apply(subst));
			}
		}
		
		for (ArrayList<Atom> newBody : lp.addAtomsForFunctions(functionSet, baseBody, substTerm, allVariables)) {
			Clause newClause = new Clause(currentQuery.getHead().apply(subst), newBody).condense();
			
			if (newBody.size() > baseBody.size()) {
				res.add(newClause);
			} else {
				res.clear();
				res.add(newClause);
				break;
			}
		}

		return res;
	}
	
	public ArrayList<MarkedClause> resolve(Clause currentQuery, Variable v, Counter mark, FlattenerExtra extra) {
		ArrayList<MarkedClause> res = new ArrayList<MarkedClause>();
		
		Set<Variable> headVars = currentQuery.getHead().getVariables();
		if (headVars.contains(v)) {
			return res;
		}
		
		Set<Term> boundTerms = currentQuery.getBoundTerms();

		FunctionMap functionSet = null;
		Set<Variable> otherVariables = new HashSet<Variable>();
		Set<Constant> otherConstants = new HashSet<Constant>();

		int[] pattern = currentQuery.termIndexPattern(v);

		for (int k = 0; k < pattern.length; k++) {
			if (pattern[k] == 0) {
				continue;
			} else if (pattern[k] == 3) {
				return res;
			}
			
			Atom atom = currentQuery.getBodyAtomAt(k);

			AtomUnfolding aur = atomUnfolder.getUnfolding(atom, boundTerms, null, extra).getUnfolding();
			
			FunctionMap newFunctions = null;

			if (atom.isConcept()) {
				newFunctions = aur.getFunctionMap(0);
			} else {
				Term otherTerm = null;
			
				if (pattern[k] == 1) {
					newFunctions = aur.getFunctionMap(0);
					if (boundTerms.contains(atom.getArgument(1))) {
						otherTerm = atom.getArgument(1);
					}
				} else if (pattern[k] == 2) {
					newFunctions = aur.getFunctionMap(1);
					if (boundTerms.contains(atom.getArgument(0))) {
						otherTerm = atom.getArgument(0);
					}
				}
				
				if (otherTerm != null) {
					if (otherTerm.isVariable()) {
						otherVariables.add((Variable)otherTerm);
					} else  if (otherTerm.isConstant()) {
						otherConstants.add((Constant)otherTerm);
					}
				}
			}

			if (otherConstants.size() > 1) {
				return res;
			}
			
			if (functionSet == null) {
				functionSet = newFunctions.copy();
			} else {
				functionSet.intersection(newFunctions);
			}
			
			if (functionSet.isEmpty()) {
				return res;
			}
		}
		
		if (functionSet == null) {
			return res;
		}

		Term substTerm = null;
		Substitution subst = new Substitution();
		
		Set<Variable> allVariables = currentQuery.getVariables();

		if (otherConstants.size() == 0) {
			if (otherVariables.size() > 0) {
				Iterator<Variable> iter = otherVariables.iterator();
				substTerm = iter.next();
				while (!headVars.contains(substTerm) && iter.hasNext()) {
					substTerm = iter.next();
				}
			} else {
				substTerm = Variable.getNewVariable(Variable.UNBOUND_PREFIX, allVariables);
				allVariables.add((Variable)substTerm);
			}
		} else {
			substTerm = otherConstants.iterator().next();
		}

		for (Variable vv : otherVariables) {
			if (!vv.equals(substTerm)) {
				subst.put(vv, substTerm);
			}
		}

		ArrayList<Atom> baseBody = new ArrayList<Atom>();
		for (int k = 0; k < currentQuery.bodySize(); k++) {
			if (pattern[k] == 0) {
				baseBody.add(currentQuery.getBodyAtomAt(k).apply(subst));
			}
		}
		
		ArrayList<Clause> last = new ArrayList<>();
		
		for (ArrayList<Atom> newBody : lp.addAtomsForFunctions(functionSet, baseBody, substTerm, allVariables)) {
			int s1 = newBody.size();
			Clause newClause = new Clause(currentQuery.getHead().apply(subst), newBody).condense();
			
			if (s1 == newClause.bodySize() && s1 >= baseBody.size() + 1) {
				res.add(new MarkedClause(newClause, mark.getValue()));
			} else if (newBody.size() <= baseBody.size()) {
				res.clear();
				res.add(new MarkedClause(newClause, mark.getValue()));
				last.clear();
				break;
			} else {
				last.add(newClause);
			}
		}
		
		for (Clause c : last) {
			res.add(new MarkedClause(c, mark.increaseUse()));
		}

		return res;
	}


	public static DatalogAtom normalizeAtom(Atom atom, Collection<Term> boundVars) {
		
		Predicate atomPred = atom.getPredicate();
		Predicate newPred = null;
		Atom clauseAtom = null;
		Atom unfoldAtom = null;
		Atom rootAtom = null;

		boolean u1 = false;
		boolean u2 = false;

		if (!boundVars.contains(atom.getArgument(0))) {
			u1 = true;
		}

		if (atom.isConcept()) {
			if (u1) {
				newPred = new Predicate(atomPred.getName() + "_u", 0);
				
				clauseAtom = new Atom(newPred, new Tuple());
				rootAtom = new Atom(newPred, new Tuple());
				unfoldAtom = new Atom(atomPred, new Tuple(Rapid.fixedUnboundVars[0]));
				
			} else {
//				newPred = new Predicate(atomPred.getName() + "_.", 1);
				newPred = atomPred;
				
				clauseAtom = new Atom(newPred, atom.getArguments());
				rootAtom = new Atom(newPred, new Tuple(Rapid.dummyVars.get(0)));
				unfoldAtom = new Atom(atomPred, new Tuple(Rapid.dummyVars.get(0)));
			}
		} else { 
			
			if (atom.getArguments().size() > 1 && !boundVars.contains(atom.getArgument(1))) {
				u2 = true;
			}
			
			if (u1 && u2) {
				newPred = new Predicate(atomPred.getName() + "_uu", 0);

				clauseAtom = new Atom(newPred, new Tuple());
				rootAtom = new Atom(newPred, new Tuple());
				unfoldAtom = new Atom(atomPred, new Tuple(Rapid.fixedUnboundVars[0], Rapid.fixedUnboundVars[1]));
				
			} else if (u1) {
				newPred = new Predicate(atomPred.getName() + "_u.", 1);

				clauseAtom = new Atom(newPred, new Tuple(atom.getArgument(1)));
				rootAtom = new Atom(newPred, new Tuple(Rapid.dummyVars.get(1)));
				unfoldAtom = new Atom(atomPred, new Tuple(Rapid.fixedUnboundVars[0], Rapid.dummyVars.get(1)));
				
			} else if (u2) {
				newPred = new Predicate(atomPred.getName() + "_.u", 1);

				clauseAtom = new Atom(newPred, new Tuple(atom.getArgument(0)));
				rootAtom = new Atom(newPred, new Tuple(Rapid.dummyVars.get(0)));
				unfoldAtom = new Atom(atomPred, new Tuple(Rapid.dummyVars.get(0), Rapid.fixedUnboundVars[1]));
			} else {
//				newPred = new Predicate(atomPred.getName() + "_..", 2);
				newPred = atomPred;
				
				clauseAtom = new Atom(newPred, atom.getArguments());
				rootAtom = new Atom(newPred, new Tuple(Rapid.dummyVars.get(0), Rapid.dummyVars.get(1)));
				unfoldAtom = new Atom(atomPred, new Tuple(Rapid.dummyVars.get(0), Rapid.dummyVars.get(1)));
			}
		}

		return new DatalogAtom(clauseAtom, rootAtom, unfoldAtom);
	}
	
	public static Clause adjustVariables(Clause c) {
		
		Set<Variable> vars = c.getVariables();
		
		Counter cc = new Counter(0);
		
		Substitution s = new Substitution();
		for (Variable v : c.getBoundVariables()) {
			if (v.hasUnboundPrefix()) {
				s.put(v, Variable.getNewVariable("v", vars, cc));
			}
		}
		
		return c.apply(s);
	}


}
