/*Copyright 2011, 2013 Alexandros Chortaras

 This file is part of Rapid.

 Rapid is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Rapid is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Rapid.  If not, see <http://www.gnu.org/licenses/>.*/

package edu.ntua.ivml.qa.test;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.PrintStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Locale;

import edu.ntua.ivml.qa.algorithm.ComputedRewriting;
import edu.ntua.ivml.qa.algorithm.Engine;
import edu.ntua.ivml.qa.algorithm.Stats;
import edu.ntua.ivml.qa.algorithm.rapid.dllite.DLRapid;
import edu.ntua.ivml.qa.algorithm.rapid.elhi.ETRapid;
import edu.ntua.ivml.common.lp.Clause;
import edu.ntua.ivml.common.lp.ClauseParser;
import edu.ntua.ivml.qa.lp.theory.FCTheory;
import edu.ntua.ivml.common.dl.LoadedOntology;
import edu.ntua.ivml.qa.utils.Utils;

public class RapidTest {

	private static FCTheory lp;
	private static ArrayList<Clause> queries; 

	private static PrintStream out = System.out;
	
	private static LoadedOntology ontRef;
	
	private static NumberFormat nf = NumberFormat.getInstance(Locale.US);
	
	static {
		nf.setMinimumFractionDigits(2);
		nf.setMaximumFractionDigits(2);
		nf.setGroupingUsed(false);
	}

	public static void main(String[] args) throws Exception {

		if (args.length < 3) {
			System.out.println("Rapid version 1.0");
			System.out.println("Syntax : RapidTest mode[DU/DD/EU/ED] ontologyURL queryFile outputFile");
			System.out.println("                        DU : unfolding for DL-Lite only with optimized subsumption checking");
			System.out.println("                        DD : unfolded datalog rewriting for DL-Lite only)");
			System.out.println("                        ED : datalog rewriting for ELHI");
			System.out.println();
			return;
		}

		String mode = args[0];
		String ontology = args[1];
		String query = args[2];
		
		String ontologyFile = "file:/" + ontology;  
		String queryFile = query;
		
		Engine engine = null;
		
		if (mode.equals("DU")) { 
			engine = DLRapid.createFastUnfoldRapid();
		} else if (mode.equals("DD")) {
			engine = DLRapid.createDatalogRapid();
		} else if (mode.equals("ED")) {
			engine = ETRapid.createDatalogRapid();
		}

		loadData(ontologyFile, queryFile);
		
		run(engine);

	}
	
	private static void run(Engine engine) throws Exception {
		if (engine == null) {
			return;
		}

		ArrayList<Clause>[] rewritings = new ArrayList[queries.size()];
		 
		String engineName = engine.getEngineName(); 
		
		if (out != null) {
			out.println();
			out.println(engineName);
		}
		
		engine.importOntology(ontRef, true);
		lp = engine.getCurrentTheory();
		
		System.out.println("Logic program size : " + lp.getClauses().size());
		
//		Utils.printObjectArrayList(System.out, " PROGRAM (" + lp.getClauses().size() + ")", lp.getSortedClauses());
		
		for (int j = 0; j < queries.size(); j++) {
			Clause query = queries.get(j);

			query.reset();
		 	
			System.out.println("Query " + (j + 1) + " : " + query);

			Stats st = new Stats();
			ComputedRewriting res = engine.computeRewritings(query, true);

	 		rewritings[j] = res.getFilteredRewritings();
		 	
		 	Stats rst = res.getStatistics();
		 	
		 	st.addIterationTimes(rst.qrewriteTime, rst.qcheckTime, rst.qmatchTime);
		 	st.finishIteration(res.getAllComputedRewritings().size(), res.getFilteredRewritings().size(), 0);
		 	
			if (out != null) {
				System.out.println();

				out.println("\tRewritten in " + st.qrewriteTime + " + " + st.qcheckTime + " (" + (st.qtotalTime) + ") ms. ");
				out.println("\tRewritings: " + st.qrewriteSize + " - " + st.qfinalSize);
				out.println();
			}

		 	
		 	Utils.printObjectArrayList(out, engineName + " - NON REDUNDANT ", rewritings[j]);
		}

		System.out.println("FINISHED.");
	}
	

	public static void loadData(String ontologyFile, String queryFile) throws Exception {
		ontRef = LoadedOntology.createFromPath(ontologyFile, true);
		
		if (ontRef != null) {
			System.out.println("Ontology : " + ontologyFile);
			System.out.println("Query    : " + queryFile);
			System.out.println();
		
			System.out.println("Ontology axioms : " + ontRef.getOntology().getLogicalAxiomCount());
		}
		
		loadQueries(queryFile);
	}
	
	private static void loadQueries(String queryFile) throws Exception {
		queries = new ArrayList<Clause>();

		ClauseParser cp = new ClauseParser(); 

		BufferedReader qbf = new BufferedReader(new FileReader(queryFile));
		
		String qline;
		while ((qline = qbf.readLine()) != null) {
			if (qline.startsWith("%")) {
				continue;
			}
			queries.add(cp.parseClause(qline));
		}
	}
	
	
}


